﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Drawing;
using System.Threading.Tasks;
using System.Runtime.InteropServices;

namespace DirectShowDemo
{
    public class Camera : IDisposable
    {
        protected readonly CameraMethods _cameraMethods;
        protected object _nativeInterface = null;
        protected readonly object _bitmapLock = new object();
        protected readonly int _index;
        protected readonly string _name;
        protected Bitmap _bitmap;
        protected DateTime _dtLastCap = DateTime.MinValue;
        protected int _fpslimit = -1;
        protected int _height = 640;
        protected double _timeBehind;
        protected double _timeBetweenFrames;
        protected int _width = 480;

        public Camera(CameraMethods cameraMethods, object nativeInterface, string name, int index)
        {
            _cameraMethods = cameraMethods;
            _name = name;
            _index = index;
            _nativeInterface = nativeInterface;
            _cameraMethods.OnImageCapture += CaptureCallbackProc;
        }
        void IDisposable.Dispose()
        {
            if (_nativeInterface != null)
            {
                Marshal.ReleaseComObject(_nativeInterface);
                _nativeInterface = null;
            }
        }
        public void Dispose()
        {
            StopCapture();
        }
        public string Name
        {
            get { return _name; }
        }

        public object NativeInterface
        {
            get { return _nativeInterface; }
        }

        /// <summary>
        /// Determines the width of the image captured
        /// </summary>
        public int CaptureWidth
        {
            get
            {
                return _width;
            }
            set
            {
                _width = value;
            }
        }

        /// <summary>
        /// Defines the height of the image captured
        /// </summary>
        public int CaptureHeight
        {
            get
            {
                return _height;
            }
            set
            {
                _height = value;
            }
        }
        /// <summary>
        /// Returns the last image acquired from the camera
        /// </summary>
        /// <returns>A bitmap of the last image acquired from the camera</returns>
        public Bitmap GetCurrentImage()
        {
            Bitmap b = null;
            lock (_bitmapLock)
            {
                if (_bitmap == null)
                {
                    return null;
                }

                b = (Bitmap)_bitmap.Clone();
            }

            return b;
        }

        public void ShowPropertiesDialog()
        {
            _cameraMethods.DisplayCameraPropertiesDialog(_index);
        }

        public void GetCameraInfo()
        {
            _cameraMethods.GetCameraInfo(_index);
        }

        /// <summary>
        /// Event fired when an image from the camera is captured
        /// </summary>
        public event EventHandler<CameraEventArgs> OnImageCaptured;

        /// <summary>
        /// Returns the camera name as the ToString implementation
        /// </summary>
        /// <returns>The name of the camera</returns>
        public override string ToString()
        {
            return _name;
        }

        internal virtual void StartCapture()
        {
            _cameraMethods.StartCamera(this, _index, ref _width, ref _height);
        }

        internal void StopCapture()
        {
            _cameraMethods.StopCamera();
        }

        /// <summary>
        /// Here is where the images come in as they are collected, as fast as they can and on a background thread
        /// </summary>
        private void CaptureCallbackProc(int dataSize, byte[] data)
        {
            GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);
            int scan0 = (int)handle.AddrOfPinnedObject();
            byte[] datacopy = new byte[dataSize];
            Marshal.Copy((IntPtr)scan0, datacopy, 0, dataSize);

            // Now you can free the handle
            handle.Free();

            // pass it to the camera for its events and processing
            ImageCaptured(datacopy, new Size(_width, _height));
        }

        public void ImageCaptured(byte[] data, Size sz)
        {
            if (OnImageCaptured != null)
            {
                OnImageCaptured.Invoke(this, new CameraEventArgs(data, sz));
            }
        }
    }

    /// <summary>
    /// Camera specific EventArgs that provides the Image being captured
    /// </summary>
    public class CameraEventArgs : EventArgs
    {
        private readonly Size _imageSize;
        private readonly byte[] _image;

        internal CameraEventArgs(byte[] data, Size sz)
        {
            _image = data;
            _imageSize = sz;
        }

        /// <summary>
        /// Current Camera Image
        /// </summary>
        public byte[] Image
        {
            get { return _image; }
        }

        public Size ImageSize
        {
            get { return _imageSize; }
        }
    }
}
