﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Runtime.InteropServices;
using System.Runtime.ExceptionServices;
using System.Drawing;
using System.Windows.Forms;

namespace DirectShowDemo
{
    public class CameraMethods
    {
        // Cameras
        static private int _videocamLibCount = 0;
        static private object _videocamLock = new object();
        private List<Camera> _cameras = null;
        //needed to avoid garbage collection problem
        private NativeMethods.CaptureCallbackProc _nativeCallback;
        private Camera _currentCamera = null;
        private int _camWidth = 0;
        private int _camHeight = 0;
        private int _camstride = 0;

        public CameraMethods()
        {
            lock (_videocamLock)
            {
                if (_videocamLibCount == 0)
                {
                    if (NativeMethods.VideoCamInitialize() != 0)
                    {
                        throw new ApplicationException("Unable to initialize the camera API");
                    }
                }

                _videocamLibCount++;
                RefreshCameraList();
            }
        }

        public void RefreshCameraList()
        {
            int count = 0;

            Cleanup();
            _cameras = new List<Camera>();

            _cameras.Add(new VideoFile(this, null, "Video File", 0));

            if (NativeMethods.VideoCamRefreshCameraList(ref count) == 0)
            {
                for (int n = 0; n < count; n++)
                {
                    IntPtr name = IntPtr.Zero;
                    object nativeInterface = null;
                    if (NativeMethods.VideoCamGetCameraDetails(n, out nativeInterface, out name) == 0)
                    {

                        Camera cam = new Camera(this, nativeInterface, Marshal.PtrToStringBSTR(name), n + 1);
                        _cameras.Add(cam);
                    }

                    Marshal.FreeBSTR(name);
                }
            }
        }

        public int ActiveCameraIndex
        {
            get
            {
                if ((_currentCamera != null) && (_cameras != null))
                {
                    return _cameras.IndexOf(_currentCamera);
                }
                return -1;
            }
        }
        public int Count
        {
            get
            {
                if (_cameras != null)
                {
                    return _cameras.Count;
                }
                return 0;
            }
            set
            {
            }
        }

        public event NativeMethods.CaptureCallbackProc OnImageCapture;

        public int Cleanup()
        {
            if (_currentCamera != null)
            {
                NativeMethods.VideoCamStopCamera();
                _currentCamera = null;
            }

            if (_cameras != null)
            {
                foreach (Camera cam in _cameras)
                {
                    cam.Dispose();
                }
                _cameras.Clear();
            }
            return 0;
        }
        public Camera GetCamera(int camIndex)
        {
            if ((_cameras != null) && (camIndex < _cameras.Count))
            {
                return _cameras[camIndex];
            }
            return null;
        }
        public void DisplayCameraPropertiesDialog(int camIndex)
        {
            if ((_cameras != null) && (_cameras.Count > camIndex))
            {
                if (_cameras[camIndex] is VideoFile)
                {
                    OpenFileDialog dlg = new OpenFileDialog();
                    dlg.Filter = "All Files (*.*)|*.*";
                    if (dlg.ShowDialog() == DialogResult.OK)
                    {
                        (_cameras[camIndex] as VideoFile).FileName = dlg.FileName;
                    }
                    else
                    {
                        throw new Exception("You must select one file");
                    }
                }
                else
                {
                    NativeMethods.VideoCamDisplayCameraPropertiesDialog(_cameras[camIndex].NativeInterface, new IntPtr(0));
                }
            }
        }
        public void Dispose()
        {
            Cleanup();

            lock (_videocamLock)
            {
                _videocamLibCount--;
                if (_videocamLibCount <= 0)
                {
                    if (NativeMethods.VideoCamCleanup() != 0)
                    {
                        throw new ApplicationException("Unable to cleanup the webcam API");  // TODO fix exception
                    }
                }
            }
        }
        [HandleProcessCorruptedStateExceptions]
        protected virtual void Dispose(bool A_0)
        {
            Dispose();
        }

        public CameraInfo GetCameraInfo(int camIndex)
        {
            CameraInfo cinfo = new CameraInfo();
            cinfo.index = camIndex;
            cinfo.name = "";
            if ((_cameras != null) && (camIndex < _cameras.Count))
            {
                cinfo.name = _cameras[camIndex].Name;
            }
            return cinfo;
        }
        protected void raise_OnImageCapture(int value0, byte[] value1)
        {
            if (OnImageCapture != null)
            {
                OnImageCapture(value0, value1);
            }
        }
        public void StartCamera(Camera cam, int camIndex, ref int width, ref int height)
        {
            if (_currentCamera != null)
            {
                NativeMethods.VideoCamStopCamera();
                _currentCamera = null;
            }

            for (int i = 0; i < 2; i++)
            {
                if ((_cameras != null) && (camIndex < _cameras.Count))
                {
                    if (_cameras[camIndex].NativeInterface == null)
                    {
                        throw new InvalidComObjectException("The camera has been disposed");
                    }
                    _nativeCallback = new NativeMethods.CaptureCallbackProc(CaptureCallbackProc);
                    int rc = NativeMethods.VideoCamStartCamera(_cameras[camIndex].NativeInterface,
                        _nativeCallback,
                        _cameras[camIndex].CaptureWidth,
                        _cameras[camIndex].CaptureHeight,
                        ref _camWidth,
                        ref _camHeight,
                        ref _camstride);
                    if (rc == 0)
                    {
                        _currentCamera = _cameras[camIndex];
                        _currentCamera.CaptureWidth = _camWidth;
                        _currentCamera.CaptureHeight = _camHeight;
                        return;
                    }
                    else
                    {
                        if (i > 0)
                        {
                            throw new ApplicationException(string.Format("Unable to start camera. rc={0}", rc)); // TODO make a better exception
                        }
                    }
                }
            }
        }
        public void StartVideoFile(VideoFile cam, int camIndex, ref int width, ref int height)
        {
            if (_currentCamera != null)
            {
                NativeMethods.VideoCamStopCamera();
                _currentCamera = null;
            }

            for (int i = 0; i < 2; i++)
            {
                if ((_cameras != null) && (camIndex < _cameras.Count))
                {
                    _nativeCallback = new NativeMethods.CaptureCallbackProc(CaptureCallbackProc);
                    int rc = NativeMethods.VideoCamStartVideoFile(new StringBuilder(cam.FileName),
                        _nativeCallback,
                        ref _camWidth,
                        ref _camHeight,
                        ref _camstride);
                    if (rc == 0)
                    {
                        _currentCamera = _cameras[camIndex];
                        _currentCamera.CaptureWidth = _camWidth;
                        _currentCamera.CaptureHeight = _camHeight;
                        return;
                    }
                    else
                    {
                        if (i > 0)
                        {
                            throw new ApplicationException(string.Format("Unable to start camera. rc={0}", rc)); // TODO make a better exception
                        }
                    }
                }
            }
        }
        public void StopCamera()
        {
            if (_currentCamera != null)
            {
                NativeMethods.VideoCamStopCamera();
                _currentCamera = null;
            }
        }

        public void CaptureCallbackProc(int dataSize, byte[] data)
        {
            if (_currentCamera != null)
            {
                // Do the magic to create a bitmap
                GCHandle handle = GCHandle.Alloc(data, GCHandleType.Pinned);
                int scan0 = (int)handle.AddrOfPinnedObject();
                byte[] datacopy = new byte[dataSize];
                Marshal.Copy((IntPtr)scan0, datacopy, 0, dataSize);

                // Now you can free the handle
                handle.Free();

                // pass it to the camera for its events and processing
                _currentCamera.ImageCaptured(datacopy, new Size(_camWidth, _camHeight));
            }
        }
    }
    public class CameraInfo
    {
        public int index { get; set; }
        public string name { get; set; }
    }
}
